@@ -4,9 +4,10 @@ use std::collections::HashMap;
 use std::ffi::OsString;
 use std::fmt::{Display, Formatter};
 use std::io;
-use std::io::{stdin, BufRead, BufReader};
+use std::io::{stdin, BufRead, BufReader, ErrorKind};
 use std::path::PathBuf;
 use std::str::FromStr;
+use std::sync::Arc;
 
 use chrono::{DateTime, FixedOffset, Local};
 use clap::AppSettings;
@@ -17,6 +18,8 @@ use crate::path::Path;
 use crate::pattern::{Pattern, PatternError, PatternOpts};
 use crate::selector::PathSelector;
 use crate::transform::Transform;
+use crate::FileGroupFilter;
+use crate::Replication::{Overreplicated, Underreplicated};
 
 #[derive(Debug, Clone, Copy)]
 pub enum OutputFormat {
@@ -71,7 +74,7 @@ fn parse_date_time(s: &str) -> Result<DateTime<FixedOffset>, String> {
             let local_offset = *Local::now().offset();
             Ok(DateTime::from_utc(dt, local_offset))
         }
-        Err(e) => Err(format!("Failed to parse {} as date: {}", s, e.to_string())),
+        Err(e) => Err(format!("Failed to parse {} as date: {}", s, e)),
     }
 }
 
@@ -263,6 +266,10 @@ pub struct GroupConfig {
       verbatim_doc_comment)]
     pub threads: Vec<(OsString, Parallelism)>,
 
+    /// Base directory to use when resolving relative input paths.
+    #[structopt(long, parse(from_os_str), default_value("."))]
+    pub base_dir: PathBuf,
+
     /// A list of input paths.
     ///
     /// Accepts files and directories.
@@ -336,6 +343,23 @@ impl GroupConfig {
             .exclude_paths(exclude_paths?))
     }
 
+    pub fn group_filter(&self) -> FileGroupFilter {
+        FileGroupFilter {
+            replication: if self.unique {
+                Underreplicated(2)
+            } else if let Some(rf) = self.rf_under {
+                Underreplicated(rf)
+            } else {
+                Overreplicated(self.rf_over())
+            },
+            root_paths: if self.isolate {
+                self.input_paths().collect()
+            } else {
+                vec![]
+            },
+        }
+    }
+
     pub fn rf_over(&self) -> usize {
         // don't prune small groups if:
         // - there is transformation defined
@@ -367,17 +391,43 @@ impl GroupConfig {
         }
     }
 
-    /// Returns an iterator over the input paths.
+    /// Makes the base directory absolute.
+    /// Returns error if the base directory does not exist.
+    pub fn resolve_base_dir(&mut self) -> io::Result<&PathBuf> {
+        if self.base_dir.is_relative() {
+            self.base_dir = std::env::current_dir()?.join(&self.base_dir)
+        }
+        if !self.base_dir.is_dir() {
+            return Err(io::Error::new(
+                ErrorKind::NotFound,
+                format!("Directory not found: {}", self.base_dir.display()),
+            ));
+        }
+        self.base_dir = self.base_dir.canonicalize()?;
+        Ok(&self.base_dir)
+    }
+
+    /// Returns an iterator over the absolute input paths.
     /// Input paths may be provided as arguments or from standard input.
     pub fn input_paths(&self) -> Box<dyn Iterator<Item = Path> + Send> {
+        let base_dir = Arc::new(if self.base_dir.as_os_str().is_empty() {
+            Path::from(".")
+        } else {
+            Path::from(&self.base_dir)
+        });
         if self.stdin {
             Box::new(
                 BufReader::new(stdin())
                     .lines()
-                    .map(|s| Path::from(s.unwrap().as_str())),
+                    .map(move |s| base_dir.resolve(Path::from(s.unwrap().as_str()))),
             )
         } else {
-            Box::new(self.paths.clone().into_iter())
+            Box::new(
+                self.paths
+                    .clone()
+                    .into_iter()
+                    .map(move |p| base_dir.resolve(p)),
+            )
         }
     }
 
@@ -501,6 +551,16 @@ pub struct DedupeConfig {
     /// Keeps files with paths matching any given patterns untouched.
     #[structopt(long = "keep-path", value_name = "pattern")]
     pub keep_path_patterns: Vec<Pattern>,
+
+    /// Specifies a list of path prefixes.
+    /// If non-empty, all duplicates having the same path prefix (root) are treated as one.
+    /// This also means that the files sharing the same root can be either all
+    /// dropped or all retained.
+    ///
+    /// By default, it is set to the input paths given as arguments to the earlier
+    /// `fclones group` command, if `--isolate` option was present.
+    #[structopt(long = "isolate", value_name = "path", parse(from_os_str))]
+    pub isolated_roots: Vec<Path>,
 }
 
 #[derive(Debug, StructOpt)]
