@@ -21,9 +21,10 @@ use crate::files::FileLen;
 use crate::lock::FileLock;
 use crate::log::Log;
 use crate::path::Path;
-use crate::util::fallible_sort_by_key;
-use crate::{Error, FileGroup, TIMESTAMP_FMT};
+use crate::util::{max_result, min_result, try_sort_by_key};
+use crate::{AsPath, Error, FileGroup, FileSubGroup, TIMESTAMP_FMT};
 use std::collections::HashMap;
+use std::time::SystemTime;
 
 /// Defines what to do with redundant files
 #[derive(Clone, PartialEq, Eq)]
@@ -72,6 +73,12 @@ impl FileMetadata {
     }
 }
 
+impl AsPath for FileMetadata {
+    fn path(&self) -> &Path {
+        &self.path
+    }
+}
+
 impl Display for FileMetadata {
     fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
         f.pad(format!("{}", self.path).as_str())
@@ -470,57 +477,6 @@ fn was_modified(files: &[FileMetadata], after: DateTime<FixedOffset>, log: &Log)
     result
 }
 
-/// Sort files so that files with highest priority (newest, most recently updated,
-/// recently accessed, etc) are sorted last.
-/// In cases when metadata of a file cannot be accessed, an error message is pushed
-/// in the result vector and such file is placed at the beginning of the list.
-fn sort_by_priority(files: &mut Vec<FileMetadata>, priority: &Priority) -> Vec<Error> {
-    let errors = match priority {
-        Priority::Newest => fallible_sort_by_key(files, |m| {
-            m.metadata
-                .created()
-                .map_err(|e| format!("Failed to read creation time of file {}: {}", m.path, e))
-        }),
-        Priority::Oldest => fallible_sort_by_key(files, |m| {
-            m.metadata
-                .created()
-                .map(Reverse)
-                .map_err(|e| format!("Failed to read creation time of file {}: {}", m.path, e))
-        }),
-        Priority::MostRecentlyModified => fallible_sort_by_key(files, |m| {
-            m.metadata
-                .modified()
-                .map_err(|e| format!("Failed to read modification time of file {}: {}", m.path, e))
-        }),
-        Priority::LeastRecentlyModified => fallible_sort_by_key(files, |m| {
-            m.metadata
-                .modified()
-                .map(Reverse)
-                .map_err(|e| format!("Failed to read modification time of file {}: {}", m.path, e))
-        }),
-        Priority::MostRecentlyAccessed => fallible_sort_by_key(files, |m| {
-            m.metadata
-                .accessed()
-                .map_err(|e| format!("Failed to read access time of file {}: {}", m.path, e))
-        }),
-        Priority::LeastRecentlyAccessed => fallible_sort_by_key(files, |m| {
-            m.metadata
-                .accessed()
-                .map(Reverse)
-                .map_err(|e| format!("Failed to read access time of file {}: {}", m.path, e))
-        }),
-        Priority::MostNested => {
-            files.sort_by_key(|m| m.path.component_count());
-            vec![]
-        }
-        Priority::LeastNested => {
-            files.sort_by_key(|m| Reverse(m.path.component_count()));
-            vec![]
-        }
-    };
-    errors.into_iter().map(Error::from).collect()
-}
-
 /// Returns true if given path matches any of the `keep` patterns
 fn should_keep(path: &Path, config: &DedupeConfig) -> bool {
     let matches_any_name = config
@@ -564,6 +520,89 @@ fn may_drop(path: &Path, config: &DedupeConfig) -> bool {
         || matches_any_path()
 }
 
+impl FileSubGroup<FileMetadata> {
+    /// Returns the time of the earliest creation of a file in the subgroup
+    pub fn created(&self) -> Result<SystemTime, Error> {
+        Ok(min_result(self.files.iter().map(|f| {
+            f.metadata
+                .created()
+                .map_err(|e| format!("Failed to read creation time of file {}: {}", f.path, e))
+        }))?
+        .unwrap())
+    }
+
+    /// Returns the time of the latest modification of a file in the subgroup
+    pub fn modified(&self) -> Result<SystemTime, Error> {
+        Ok(max_result(self.files.iter().map(|f| {
+            f.metadata
+                .modified()
+                .map_err(|e| format!("Failed to read modification time of file {}: {}", f.path, e))
+        }))?
+        .unwrap())
+    }
+
+    /// Returns the time of the latest access of a file in the subgroup
+    pub fn accessed(&self) -> Result<SystemTime, Error> {
+        Ok(max_result(self.files.iter().map(|f| {
+            f.metadata
+                .accessed()
+                .map_err(|e| format!("Failed to read access time of file {}: {}", f.path, e))
+        }))?
+        .unwrap())
+    }
+
+    /// Returns true if any of the files in the subgroup must be kept
+    pub fn should_keep(&self, config: &DedupeConfig) -> bool {
+        self.files.iter().any(|f| should_keep(&f.path, config))
+    }
+
+    /// Returns true if all files in the subgroup can be dropped
+    pub fn may_drop(&self, config: &DedupeConfig) -> bool {
+        self.files.iter().all(|f| may_drop(&f.path, config))
+    }
+
+    /// Returns the number of components of the least nested path
+    pub fn min_nesting(&self) -> usize {
+        self.files
+            .iter()
+            .map(|f| f.path.component_count())
+            .min()
+            .unwrap()
+    }
+
+    /// Returns the number of components of the most nested path
+    pub fn max_nesting(&self) -> usize {
+        self.files
+            .iter()
+            .map(|f| f.path.component_count())
+            .max()
+            .unwrap()
+    }
+}
+
+/// Sort files so that files with highest priority (newest, most recently updated,
+/// recently accessed, etc) are sorted last.
+/// In cases when metadata of a file cannot be accessed, an error message is pushed
+/// in the result vector and such file is placed at the beginning of the list.
+fn sort_by_priority(files: &mut [FileSubGroup<FileMetadata>], priority: &Priority) -> Vec<Error> {
+    match priority {
+        Priority::Newest => try_sort_by_key(files, |m| m.created()),
+        Priority::Oldest => try_sort_by_key(files, |m| m.created().map(Reverse)),
+        Priority::MostRecentlyModified => try_sort_by_key(files, |m| m.modified()),
+        Priority::LeastRecentlyModified => try_sort_by_key(files, |m| m.modified().map(Reverse)),
+        Priority::MostRecentlyAccessed => try_sort_by_key(files, |m| m.accessed()),
+        Priority::LeastRecentlyAccessed => try_sort_by_key(files, |m| m.accessed().map(Reverse)),
+        Priority::MostNested => {
+            files.sort_by_key(|m| m.max_nesting());
+            vec![]
+        }
+        Priority::LeastNested => {
+            files.sort_by_key(|m| Reverse(m.min_nesting()));
+            vec![]
+        }
+    }
+}
+
 struct PartitionedFileGroup {
     to_keep: Vec<FileMetadata>,
     to_drop: Vec<FileMetadata>,
@@ -576,8 +615,8 @@ impl PartitionedFileGroup {
         let root = source_path
             .root()
             .to_string()
-            .replace("/", "")
-            .replace(":", "");
+            .replace('/', "")
+            .replace(':', "");
         let suffix = source_path.strip_root();
         if root.is_empty() {
             target_dir.join(suffix)
@@ -641,6 +680,29 @@ impl PartitionedFileGroup {
     }
 }
 
+/// Attempts to retrieve the metadata of all the files in the file group.
+/// If metadata is inaccessible for a file, a warning is emitted to the log, and None gets returned.
+fn files_metadata(group: FileGroup<Path>, log: &Log) -> Option<Vec<FileMetadata>> {
+    let mut last_error: Option<io::Error> = None;
+    let files: Vec<_> = group
+        .files
+        .into_iter()
+        .filter_map(|p| {
+            FileMetadata::new(p)
+                .map_err(|e| {
+                    log.warn(&e);
+                    last_error = Some(e);
+                })
+                .ok()
+        })
+        .collect();
+
+    match last_error {
+        Some(_) => None,
+        None => Some(files),
+    }
+}
+
 /// Partitions a group of files into files to keep and files that can be safely dropped
 /// (or linked).
 fn partition(
@@ -657,30 +719,10 @@ fn partition(
         )))
     };
 
-    // Fetch metadata of the files
-    let mut metadata_err = false;
-    let mut files: Vec<_> = group
-        .files
-        .into_iter()
-        .map(|p| {
-            FileMetadata::new(p)
-                .map_err(|e| {
-                    log.warn(e);
-                    metadata_err = true;
-                })
-                .ok()
-        })
-        .flatten()
-        .collect();
-
-    // On metadata errors, we're just ignoring the group.
-    // We must not attempt removing any files from the group
-    // because we can't reliably determine which files to remove.
-    // What if the files we couldn't access should be removed and the ones we access
-    // should remain untouched?
-    if metadata_err {
-        return error("Metadata of some files could not be obtained");
-    }
+    let mut files = match files_metadata(group, log) {
+        Some(files) => files,
+        None => return error("Metadata of some files could not be obtained"),
+    };
 
     // We don't want to remove dirs or symlinks
     files.retain(|m| {
@@ -714,13 +756,15 @@ fn partition(
         }
     }
 
+    let mut file_sub_groups = FileSubGroup::group(files, &config.isolated_roots);
+
     // Sort files to remove in user selected order.
     // The priorities at the beginning of the argument list have precedence over
     // the priorities given at the end of the argument list, therefore we're applying
     // them in reversed order.
     let mut sort_errors = Vec::new();
     for priority in config.priority.iter().rev() {
-        sort_errors.extend(sort_by_priority(&mut files, priority));
+        sort_errors.extend(sort_by_priority(&mut file_sub_groups, priority));
     }
 
     if !sort_errors.is_empty() {
@@ -730,11 +774,11 @@ fn partition(
         return error("Metadata of some files could not be read.");
     }
 
-    // Split the set of files into two sets - a set that we want to keep intact and a set
+    // Split the set of file subgroups into two sets - a set that we want to keep intact and a set
     // that we can remove or replace with links:
-    let (mut to_retain, mut to_drop): (Vec<_>, Vec<_>) = files
+    let (mut to_retain, mut to_drop): (Vec<_>, Vec<_>) = file_sub_groups
         .into_iter()
-        .partition(|m| should_keep(&m.path, config) || !may_drop(&m.path, config));
+        .partition(|m| m.should_keep(config) || !m.may_drop(config));
 
     // If the set to retain is smaller than the number of files we must keep (rf), then
     // move some higher priority files from `to_drop` and append them to `to_retain`.
@@ -744,8 +788,8 @@ fn partition(
 
     assert!(to_retain.len() >= n || to_drop.is_empty());
     Ok(PartitionedFileGroup {
-        to_keep: to_retain,
-        to_drop,
+        to_keep: to_retain.into_iter().flat_map(|g| g.files).collect(),
+        to_drop: to_drop.into_iter().flat_map(|g| g.files).collect(),
     })
 }
 
@@ -853,6 +897,7 @@ pub fn log_script(
 #[cfg(test)]
 mod test {
     use std::collections::HashSet;
+    use std::fs::create_dir;
     use std::path::PathBuf;
     use std::{thread, time};
 
@@ -1157,6 +1202,40 @@ mod test {
         })
     }
 
+    #[test]
+    fn test_partition_respects_isolated_roots() {
+        with_dir("dedupe/partition/isolated_roots", |root| {
+            let root1 = root.join("root1");
+            let root2 = root.join("root2");
+            create_dir(&root1).unwrap();
+            create_dir(&root2).unwrap();
+
+            let group1 = make_group(&root1);
+            let group2 = make_group(&root2);
+            let group = FileGroup {
+                file_len: group1.file_len,
+                file_hash: group1.file_hash,
+                files: group1.files.into_iter().chain(group2.files).collect(),
+            };
+
+            let mut config = DedupeConfig::default();
+            config.priority = vec![Priority::Oldest];
+            config.isolated_roots = vec![Path::from(&root1), Path::from(&root2)];
+
+            let p = partition(group.clone(), &config, &Log::new()).unwrap();
+            assert_eq!(p.to_drop.len(), 3);
+            assert!(p
+                .to_drop
+                .iter()
+                .all(|f| f.path.to_path_buf().starts_with(&root1)));
+            assert_eq!(p.to_keep.len(), 3);
+            assert!(p
+                .to_keep
+                .iter()
+                .all(|f| f.path.to_path_buf().starts_with(&root2)));
+        })
+    }
+
     #[test]
     fn test_run_dedupe_script() {
         with_dir("dedupe/partition/dedupe_script", |root| {
